''Propeller file mmc.spin
'
' Read MMC and SD cards. Figure out type of card, type of filesystem and
' work with FAT structure.
'
' Initialize with relatively slow Spin card access. Once card is started,
' start an assembly cog for much faster sector read.
'
' State of progress:
'   lists all root directory filenames - tested on MMC 16M and 64M, and SD 64M and 128M cards.
'   Locates and lists first lines of root dir text files, tracking
'   allocated clusters thru a FAT16 table.

  
CON    
         
   _clkmode = xinput + pll16x         ' use external clock multiplied by 16 -> 64 mHz
  _xinfreq = 4_000_000                ' external rate is 4 MHz

' One LCD display and two separate LEDs are mounted on the board:
'
'   7 SEGMENT LCD: ports 0-6. This is used in the "worm" loop.
'   LCD_DATA       port  20
'   LCD_POWER      port  21
'   LED_BOARD      port  27
                                                    

   QUIET=false
   DISPLAY=true
     
   MMCSD_CS       = 8                            ' These 4 pins connect to the card socket.
   MMCSD_CARDIN   = 9                            ' The three outs connect directly to card pins.  
   MMCSD_CLK      = 10
   MMCSD_CARDOUT  = 11                           ' cardout has 10K pullup and 220 ohm series prot res 
   
   MMCSD_HISSIZE  = 200                          ' size of an array of command response

   UNKNOWN_TYPE   = 0                 
   MMC_TYPE       = 1                            ' type of card
   SD_TYPE        = 2
   FAT12          = 1                            ' type of filesystem
   FAT16          = 2

   
OBJ

  lcd:    "Lcd"

     
VAR                                     

  byte Response[MMCSD_HISSIZE]
  long responseIndex
  byte buffer[516]              
  byte directoryEntry[32]      
 
' These are set by "card_init"  
  long cardType  
  long totalCardSectors

' these are set by "filesystem_info"
  long numberOfRootDirEntries, totalSectors, sectorsPerFAT
  long partitionStart, FATStart, dirStart, datStart
  long fileSystemType, sectorsPerCluster, numFATs   
  
' this is set by read_dir
  long currentSector 
  
  long sendGetByteArg
  long asmCardCogStarted

  byte fast

                     
PUB main | i  

  fast := true          'set false for slow (Spin) IO
  
  lcd.reset
     
  card_init(QUIET)                 ' DISPLAY -> show lots of initialization information
     
  filesystem_info(QUIET)           ' QUIET -> initialize quickly and quietly

  findFile(string("SPINFILE"))


' when in FAST mode, uncommentiong this delay call will break the program:,
' but commenting it results in fine-working FAST-mode execution.
' 
'  delay_ms(10)
  
  repeat i from dirStart to dirStart+31
    listDirSector(i)
  
  worm

    
PRI findFile(fileName) | sector, entry, index, i, j, k, fileSize, cluster, bytecount     

' quick and dirty routine which finds the location of "SPINFILE.TXT" in a directory,
' locates its data sectors, and displays the first 20 characters of all 10 sectors,
' following its allocation through 3 clusters using the FAT table.
 
  entry := 0
  
  lcd.line1
  lcd.str(string("Looking for "))
  lcd.str(fileName)
  
  sector := dirStart 
  sector := $111
  
  repeat while entry < numberOfRootDirEntries

    readSector(sector++)
    index := 0
    
    repeat j from 0 to 15
      entry++
      
      repeat i from 0 to 31     
        directoryEntry[i] := buffer[index++]
      
      if directoryEntry[0] == "S" and directoryEntry[1] == "P"

        lcd.clear
  
        lcd.line1
        repeat i from 0 to 7              
          lcd.tx(directoryEntry[i])
        lcd.tx(".")
        repeat i from 8 to 10              
          lcd.tx(directoryEntry[i])

        fileSize := directoryEntry[$1C] + (directoryEntry[$1D] << 8) + (directoryEntry[$1E] << 16) + (directoryEntry[$1F] << 24)
        cluster  :=  directoryEntry[$1A] + (directoryEntry[$1B] << 8)       

' The initial cluster may be translated directly to data sectors without using the FAT.
' It is only if the file size is greater than a cluster, typically 4 sectors, 2048 bytes,
' that we need to goto the FAT for the location of the next cluster.

        byteCount := 0

        repeat until byteCount > fileSize
        
          sector   := clusterToSector(cluster)    
                 
          repeat k from 0 to sectorsPerCluster - 1
            readSector(sector + k)

            lcd.line1
            
            repeat i from 0 to 19      
              if (byteCount++ => fileSize)
                 return
              lcd.tx(buffer[i])
                                
            byteCount += 492

            if (byteCount > fileSize)
              return
                          
            delay_ms(500)
          
' When we've accessed the last sector in this cluster, find the value of the next cluster:

          cluster := nextFATentry(cluster)

   
PUB clusterToSector(cluster)

  return ((cluster-2) * sectorsPerCluster) + datStart


PUB nextFATentry(cluster) | entryIndex, fatSector

  if fileSystemType == FAT16                     ' FAT16 contains 256 entries per FAT sector

    fatSector := fatStart + ((cluster * 2) / 512)
  
    readSector(fatSector)

    entryIndex := cluster * 2

    Result := buffer[entryIndex] + (buffer[entryIndex+1] << 8) 

  elseif  fileSystemType == FAT12  ' not yet worked out

    Result := 0

  else 
    error
    
  
PUB rootdir | sector

  sector := dirStart

  repeat until listDirSector(sector++) == false
    
  
PUB listDirSector(sector) | entry, index, i, j, status1, status2, filesize, firstCluster
'
' This is an incomplete but much more detailed directory listing.
' This routine identifies subdirectories, erased files, longfilenames, etc.

  readSector(sector)
  index := 0
    
  repeat entry from 0 to 15

    repeat i from 0 to 31     
      directoryEntry[i] := buffer[index++]
          
    status1 := directoryEntry[0]
    status2 := directoryEntry[11]
              
    lcd.clear
    lcd.line1
            
    if (status1 == 0)
      lcd.str(string("End of entries"))
      return false
       
    elseif (status1 == $E5)
      lcd.str(string("Erased "))
           
    elseif (status1 == $2E)
      lcd.str(string("DotDot entry"))
    
    elseif ((status2 & $F) == $F) and ((status1 & $40) == $40)
      lcd.str(string("LFN E  "))
      dirEntryName
 
    elseif ((status2 & $F) == $F) and ((status1 & $40) <> $40)
      lcd.str(string("LFN  "))
      dirEntryName

    elseif ((status2 & $8) == $8)
      lcd.str(string("Volume "))
      dirEntryName    
         
    elseif ((status2 & $10) == $10)
      lcd.str(string("Drctory "))
      dirEntryName 
   
' we should be left with only regular files

    else
    
      lcd.str(string("File   "))

      dirEntryName

      fileSize := directoryEntry[$1C] + (directoryEntry[$1D] << 8) + (directoryEntry[$1E] << 16) + (directoryEntry[$1F] << 24) 
      firstCluster := directoryEntry[$1A] + (directoryEntry[$1B] << 8)
      lcd.snsn2(string("Sz "), fileSize, string("Cl "), firstCluster)
                      
    delay_ms(500)

  return true

  
PUB dirEntryName | i
      
  repeat i from 0 to 10
    lcd.tx(directoryEntry[i])

  delay_ms(1000) 

               
PUB filesystem_info(verbose) |  i, r, bytesPerSector, reserved
               
  if not find_boot_partition(verbose)
    lcd.str1x(string("cant find boot sect "),r,2)
    return false
                  
' show the ASCII filesystem type (FAT12, FAT16, etc)

  if (verbose)
    lcd.line1 
    if fileSystemType == FAT12
      lcd.str(string("FS FAT12: "))
    elseif fileSystemType == FAT16
      lcd.str(string("FS FAT16: "))
    else
      lcd.str(string("FS type ? "))

    repeat i from $36 to $3D  
      lcd.tx(buffer[i])
    delay_ms(1000) 
  
' Now, for either FAT12 or FAT16, the buffer should contain the filesystem boot sector.
' Lets read a bunch of filesystem parameters.

  bytesPerSector         := buffer[$B] + (buffer[$C] << 8)
  sectorsPerCluster      := buffer[$D]
  reserved               := buffer[$E] + (buffer[$F] << 8)
  numFATs                := buffer[$10]  
  numberOfRootDirEntries := buffer[$11] + (buffer[$12] << 8) 
  totalSectors           := buffer[$13] + (buffer[$14] << 8)

  if (totalSectors == 0)
    totalSectors         := buffer[$20] + (buffer[$21] << 8) +(buffer[$22] << 16) +(buffer[$23] << 24)
  
  sectorsPerFAT          := BUFFER[$16] + (buffer[$17] << 8) 

  FATStart := partitionStart + reserved       
  dirStart := FATStart + (sectorsPerFAT * numFATs)
  datStart := dirStart + (numberOfRootDirEntries >> 4)  ' (32 chars/entry) / 512 chars/sector) -> %16 -> >>4
 

  if (verbose)
    lcd.str1x(string("bytes per sect "), bytesPerSector,4)
    lcd.str1x(string("sect pr clust "),  SectorsPerCluster,2)
    lcd.str1x(string("reserved sect "),  reserved,4)
    lcd.str1x(string("number of FAT "),  numFATs,1)
    lcd.str1x(string("rootDirEntrys "),  numberOfRootDirEntries ,4)
    lcd.str1x(string("sects per FAT "),  sectorsPerFAT,4)
    lcd.str1x(string("total sects "),    totalSectors,8)    
    lcd.str1x(string("dirStart    "),    dirStart,4)
    lcd.str1x(string("FATstart    "),    FATstart,4)
    lcd.str1x(string("datStart    "),    datStart,4)

    delay_ms(1000)
    
  return true

 
PUB find_boot_partition(verbose) | s

' Some cards do not have a partition table. Instead, sector 0 is the boot sector,
' the first sector of the filesystem partition. So first check that case.

  partitionStart := 0
  
  if is_boot_sector(verbose)
    return true

' If partition 0 is not a boot sector then it should contain partition table entries.    
' DOS may have up to 4 partitions, with filesystems in any or all of them.
' In practice, all cards tested have used only the first partition, but for
' completeness lets try them all if necessary...

' If 0 is not a boot sector, then try to select one of four remote partition table entries.
'
' Each of 4 partition table addresses are 16 bytes starting at address $1BE, $1CE, $1DE, $1EE
' Four byte address of first sector in each partition is 8 bytes into each table. 

  repeat s from $1BE to $1EE step 16
    partitionStart := buffer[s+8] + (buffer[s+9] << 8) + (buffer[s+10] << 16) + (buffer[s+11] << 24) 

    if is_boot_sector(verbose)
      return true

  lcd.str(string("FAIL find filesystem"))       

  return false
                       

PUB is_boot_sector(verbose)

  fileSystemType := UNKNOWN_TYPE

  if (verbose)    
    lcd.str1n(String("try boot sct "),partitionStart,6)
             
  readSector(partitionStart) 
 
  if (buffer[$1FE] <> $55) or (buffer[$1FF] <> $AA)  ' Look for FAT signature
    return false

' If the ASCII characters "FAT" exist, then this is the boot sector.
'                    F                      A                      T                                        
  if buffer[$36] == $46 and buffer[$37] == $41 and buffer[$38] == $54

    if buffer[$39] == $31 and buffer[$3A] == $32
      fileSystemType := FAT12
      return true

    if buffer[$39] == $31 and buffer[$3A] == $36
      fileSystemType := FAT16          
      return true

' is FAT32 ever used for cards?
    
  return false
       
                        
PUB card_init(verbose) | i,type,READ_BL_LEN,C_SIZE,C_SIZE_MULT,MULT,BLOCK_LEN,BLOCK_NR,MEM, initTest

  if (verbose)    
    lcd.str1n(String("Card init starts "),0,1)
      
  cardtype := UNKNOWN_TYPE
  responseIndex := 0
     
  dira[MMCSD_CS]  := 1
  dira[MMCSD_CARDIN] := 1
  dira[MMCSD_CLK] := 1
  dira[MMCSD_CARDOUT] := 0

  mmcsd_cs_lo                   

  repeat i from 1 to 10                          ' 80 clocks initialize MMC and SD cards
    spinSendGetByte($FF)  
                                 
  spinCommand(0,0)                               ' all cards start with command 0
  spinCloseCommand                               ' CS low during CMD0 requests SPI mode. 
          
  type := spinCommand(0,55)                      ' MMC does not recognize CMD55, and returns err code 5 

  if (type == 5)                                 ' so receipt of code 5 is how we recognize MMC.
    cardtype := MMC_TYPE
    command_1(verbose)                           ' MMC initialization is completed with CMD1
    if (verbose)
      lcd.str1n(String("MMC card type "), type, 1)             
  else
    cardtype := SD_TYPE                          ' SD initialization is completed with CMD41_55 pair
    command_4155(verbose)
    if (verbose)    
      lcd.str1n(String("SD card type "),type,1)

' Read and decode the CSD register.
' Algorithm from page 3-15 of document MMCProdMn_rev51.pdf
' "MultiMediaCard Product Manual" rev 5.1, SanDisk corp.
      
  readRegister(9,16)

' READ_BL_LEN: 4 bits from array 83:80; exponent of bytes per block
' value usually "9"
  READ_BL_LEN := $F & Response[5]
 
  if (READ_BL_LEN <> 9)   ' something is wrong unless 9 (512 byte blocks)
    lcd.str1x(string("ERR: BL not 9: "),READ_BL_LEN,2)
    return 0
    
' BLOCK_LEN = 2 (exp read_block_len)
' value usually 512 bytes per block
  BLOCK_LEN := 1 << READ_BL_LEN
  
' C_SIZE_MULT: 3 bits. 2 top bits from array 10, 1 bit from array 11.
' values 16M: 2, 64M: 4 
  C_SIZE_MULT := ((  3 & Response[9] ) << 1) + ((128 & Response[10]) << 7)

' MULT is 2 exp (c_size_mult + 2)
' values 16 for 16MB card, 64 for 64MB card
  MULT := 1 << (C_SIZE_MULT + 2)

' C_SIZE: 12 bits: 2 lo bits of byte 6, byte 7, 2 hi bits of byte 8
' values 16MB: 1963, 64MB: 1959; 256M: 3929?
  C_SIZE := ((3 & Response[6]) << 10) + (Response[7] << 2) + (($C0 & Response[8]) >> 6)

' BLOCK_NR
' values 16M: 31424, 64M: 125440, 256M: 250880
  BLOCK_NR := (C_SIZE+1) * MULT
  totalCardSectors := BLOCK_NR

' bytes: block count times block size
' values: 16M: 16089088,  64M: 64225280, 128M? 128450560 
  MEM := BLOCK_NR * BLOCK_LEN
     
  if (verbose)
    lcd.str1n(string("RD Bk len (9): "),  READ_BL_LEN, 1)
    lcd.str1n(string("Blk len: (512): "), BLOCK_LEN, 3)
    lcd.str1n(string("CSM (2, 4): "),     C_SIZE_MULT, 1) 
    lcd.str1n(string("MULT (16,64): "),   MULT, 2)
    lcd.str1n(string("S 1963,1959 "),     C_SIZE, 4)
    lcd.str1x(string("card sects "),      BLOCK_NR, 6)
    lcd.str2x(string("card bytes "),      MEM, 8)
    lcd.str1n(String("End of card_init "),0,1)
    delay_ms(1000)
    lcd.clear                                                 


  if fast
    
' By the time card_init ends, the card is ready for high-speed sector reads..

' The IO pins must either be set to input or set low in order
' for them to be shared among cogs. Lets set them as inputs.
' Then start the cog and let it warm up.

' (try different combinations of IO pin release trying to debug
' misbehaving cog problem.
' It makes no difference, whether out -> 0, or dir -> in, or both.)
 
    outa[MMCSD_CARDIN] := 0
    outa[MMCSD_CLK] := 0
    
'    dira[MMCSD_CARDIN] := 0                                       ' 1. release the pins
'    dira[MMCSD_CLK] := 0

    
    asmCardCogStarted := cognew(@sendGetByte, @sendGetByteArg)    ' 2. start the cog

    if (asmCardCogStarted < 1 or asmCardCogStarted > 7)
      lcd.str1n(string("Cog failed "), asmCardCogStarted , 4)
      error
      
    delay_ms(10)       


PUB readSector(address) | i, d, reps 

  if not fast
    return slowReadSector(address)
      
' The chips take a 31 bit byte address, but for most purposes, the sector
' address is more useful. So let the passed-in "address" refer to 512-byte
' sectors. Convert sector number to byte number before addressing card.

  mmcsd_cs_lo 

  cardSendGetByte($51)                            ' $40 (bit 5 set) and command 17 -> $51
  cardSendGetByte((address >> 15) & $FF)          ' address byte 3
  cardSendGetByte((address >> 7)  & $FF)          ' address byte 2
  cardSendGetByte((address << 1)  & $FF)          ' address byte 1
  cardSendGetByte( 0 )                            ' address byte 0
  cardSendGetByte($FF) 
        
  reps := 0
  d := cardSendGetByte($FF)
  
  repeat while (d <> $FE) and (reps++ < 50)     ' spin till $FE arrives. 
    d := cardSendGetByte($FF)                  ' 4 usually works. Lets set to 50 for wide margin. 
  
  if (d <> $FE)  
    lcd.str1x(String("Err rd addr "),address,6)
    error
  
  repeat i from 0 to 513                       ' retrieve 512 data bytes plus 2 CRC bytes             
    buffer[i] := cardSendGetByte($FF)
    
  cardSendGetByte($FF)
  cardSendGetByte($FF)

  mmcsd_cs_hi

  cardSendGetByte($FF)
  cardSendGetByte($FF)

  currentSector := address

  
PUB cardSendGetByte(toSend) | loops        

' Set bit $100 on byte to send to card, then wait for bit $200 to be set,
' at which point the result is ready to use.
  
  sendGetByteArg := toSend | $100                                 ' send byte to cog

  loops := 0  
  repeat while ((sendGetByteArg & $200) == 0) and (loops++ < 10)  ' wait till cog is done

  if (loops > 9)
    lcd.str1x(string("COG timeout "), sendGetByteArg , 3)
    error
     
  Result := sendGetByteArg & $FF                                   ' return result byte


PUB slowReadSector(address) | i, resul
       
'' Read a sector using entirely Spin commands.
''
'' timing tests:
''   One sendgetbit    8752 clox   (136 microseconds)
''   One sendgetByte: 71280 clox   (1.114 milliseconds) 
''   One full cmd:   445120 clox   (6.955 milliseconds) (about 7 sendGetBytes) (read sector)
''   512 reads     37322032 clox   (583.157 milliseconds) (read back the 512 data bytes)
''   one second    64000000
'
''   So the 512 reads does take about a half second.
''   And 1/512 of that does work out to about 72000 clox.
''   And one-eighth of that does work out to about 9000 clox.
''
'' So looks like no better speed is possible without going to assembly...
''
  resul := spinCommand(address, 17)
  
  if (resul <> $FE)
    spinCloseCommand
    lcd.str1n(String("Error Reading "),address,6)
    error
  
  repeat i from 0 to 513                    ' 512 data bytes plus 2 CRC bytes             
    buffer[i] := spinSendGetByte($FF)
  spinCloseCommand

  currentSector := address

  
PUB readRegister(register, length) | i

  spinCommand(0,register)

  responseIndex := 0
    
' Registers vary from 4 to 16 bytes.
' Read more to read 2 byte CRC and two $FF to make sure we're at the end.

  repeat i from 0 to length + 4
    Response[responseIndex++] := spinSendGetByte($FF)

  spinCloseCommand


PRI command_4155(verbose) | a,b

' SD cards need command 41 then 55 repeated until command 41 returns 0

  spinCommand(0,41)

  a := 0
  b := $FF
  
  repeat while (a++ < 50) and (b > 0)
    
    spinCommand(0,55)
    b :=   spinCommand(0,41)
    delay_ms(2)

  if (verbose)    
    if (b == 0)
      lcd.str1(String("SD initialized     "))
      delay_ms(500)
    else
      lcd.str1(String("unknown card?"))
      delay_ms(500)


PUB command_1(verbose) | a,b

' MMC cards need command 1 repeated until it returns 0

  a := 0
  b := $FF
    
  repeat while (a++ < 20 and b <> 0)

    b := spinCommand(0,1)

  if (verbose)
    if (b == 0)
      lcd.str(String("MMC initialized     "))
      delay_ms(500)
    else
      lcd.str(String("MMC not ok?"))
      delay_ms(500)
        
          
PRI spinCommand(address, command) | addr0, addr1, addr2, addr3, t, reps

' execute command and return result.
' Most commands return $FE.

' The chips take a 31 bit byte address, but for most purposes, the sector
' address is more useful. So let the passed-in "address" refer to 512-byte
' sectors. Convert sector number to byte number before addressing card.

  addr0 := 0
  addr1 := (address << 1)  & $FF
  addr2 := (address >> 7)  & $FF
  addr3 := (address >> 15) & $FF

  mmcsd_cs_lo 
    
  spinSendGetByte(command | $40)
  spinSendGetByte(addr3)
  spinSendGetByte(addr2)
  spinSendGetByte(addr1)
  spinSendGetByte(addr0)
  spinSendGetByte($95)                           ' $95 is CRC required for very first command
  
' If the command is read register(9) or read sector (17) spin till $FE arrives.

  if ((command == 9) or (command == 17))          
    reps := 0    
    repeat while (t <> $FE) and (reps++ < 100)  ' timeout value: 20 fails 30 ok 100 ok
      t := spinSendGetByte($FF)          
    return t & $FF
           
' The remaining commands: (0,1,5,41,...) get result byte, then sendget one more byte.

  spinSendGetByte($FF)
  t := spinSendGetByte($FF)
  spinSendGetByte($FF)

  return t

  
PRI spinCloseCommand

  spinSendGetByte($FF)
  spinSendGetByte($FF)
  mmcsd_cs_hi  
  spinSendGetByte($FF)
  spinSendGetByte($FF)
  
  
PRI spinSendGetByte(byteToSend) | i, in                       
  
  in := 0
  
  repeat i from 1 to 8              ' both send and receive 8 bits

    in <<= 1                        ' pre-shift arriving byte (dont want to shift last bit)
    
    if ((byteToSend & $80) == 0)    ' send high-bit first
      mmcsd_dt_lo
    else
      mmcsd_dt_hi

    in |= mmcsd_card_out            ' also receive hi bit first and tack onto end of result.  
      
    byteToSend <<= 1                ' rotate outgoing byte.
    
    mmcsd_ck_hi
    mmcsd_ck_lo
  
  Result := in

'---------------------------------------------------------------
' Below are all card hardware bit assignments

PRI mmcsd_card_out

  if ina[MMCSD_CARDOUT] == 1               ' "OUT" refers to the card's point of view,
    return 1                               '  so CARDOUT is IN to the Propeller.
  else
    return 0
      
PRI mmcsd_dt_hi

  outa[MMCSD_CARDIN] := 1                  ' And "IN" to the card is OUT from the Propeller.
  
PRI mmcsd_dt_lo

  outa[MMCSD_CARDIN] := 0
    
PRI mmcsd_ck_hi

  outa[MMCSD_CLK] := 1
  
PRI mmcsd_ck_lo

  outa[MMCSD_CLK] := 0

PRI mmcsd_cs_lo

  outa[MMCSD_CS] := 0
      
PRI mmcsd_cs_hi
  
  outa[MMCSD_CS] := 1

'---------------------------------------------------------------
' utilities

PUB error

' display error message then spin endlessly

  lcd.str2(string("Halted."))
  worm
  
         
PUB worm | i, del
'
' display a worm-like pattern of looping segments of a 7-segment LED display.
' This endless loop is used all over the place as a "job finished" indicator.

  del := 8_000_000

  repeat i from 0 to 6
    dira[i] := 1
      
  repeat

    outa[6] := 0
    outa[2] := 1
    waitcnt(cnt + del)
    outa[1] := 0
    outa[3] := 1
    waitcnt(cnt + del)
    outa[2] := 0
    outa[6] := 1
    waitcnt(cnt + del)
    outa[3] := 0
    outa[0] := 1
    waitcnt(cnt + del)
    outa[6] := 0
    outa[5] := 1
    waitcnt(cnt + del)
    outa[0] := 0
    outa[4] := 1
    waitcnt(cnt + del)
    outa[5] := 0
    outa[6] := 1
    waitcnt(cnt + del)
    outa[4] := 0
    outa[1] := 1
    waitcnt(cnt + del)    

   
PUB delay_ms(ms)

  waitcnt(cnt + (ms * 64_000))

 
PUB delay_us(us)

  if (us < 24)
    us := 24
    
  waitcnt(cnt + us * 64)

        
DAT

'-------------------------------------------------------------------  
' Asm notes
' 1. each routine must start with "org"
' 2. variables for each routine need to be located after its code,
'    but within the scope of its "org".
' 3. var declarations: "LONG" must be declared before "RES".
'   
'-------------------------------------------------------------------  
' MMC/SD card sendGetByte
'
' Communicate with Spin program through a single long.
' Loop while reading single long from hub.
' Do nothing while bit $100 is clear.
' When bit $100 is set then a byte transfer is being requested.
' Send the low 8 bits to the card and get a byte from the card.
' Return a long to the hub containing the card return byte,
'   with bit $100 clear and bit $200 set.
' Then go back to looping.
 
        org

sendGetByte  

        mov       argAddress, par                ' grab hub address of IO long
                              
        mov       inPinMask,  #1                 ' prepare card-in pin mask
        shl       inPinMask, L_MMCSD_CARDIN

        mov       clkPinMask, #1                 ' prepare card clock pinmask
        shl       clkPinMask, L_MMCSD_CLK

        mov       outPinMask, #1                 ' prepare card-out pinmask
        shl       outPinMask, L_MMCSD_CARDOUT    ' (pin should be input already)
        
        and       outa, !inPinMask               ' prepare for output to be low
        or        dira, inPinMask                ' Set pin to output
        and       outa, !clkPinMask              ' prepare for output to be low
        or        dira, clkPinMask               ' Set pin to output 
 
        mov       readyMask, #1                  ' prepare data-ready bitmask
        shl       readyMask, #8
        
        mov       doneMask, #1                   ' prepare data-done bitmask
        shl       doneMask, #9

toploop
     
        rdlong    argData, argAddress
        
        mov       bytToSend, argData       
        and       bytToSend, #$FF
        
        mov       dataReadyFlag, argData
        and       dataReadyFlag, readyMask wz
    
 if_z   jmp       #toploop                       ' loop if no data to process
 
        mov       sendIndex, #8                  ' send 8 bits to card and
        mov       gotByte, #0                    ' receive 8 bits in return.

:loop   shl       gotByte, #1

        mov       temp, bytToSend
        and       temp, #$80 
        test      temp, #$80 wz           
  if_nz or        outa, inPinMask                ' when hi bit set, make pin hi 
  if_z  and       outa, !inPinMask               ' when hi bit NOT set, make pin lo

        mov       temp, ina
        and       temp, outPinMask
        test      temp, outPinMask wz            ' If incoming data is high, set the result bit            
  if_nz or        gotByte, #1                    ' when input pin set, then set data bit  

        shl       bytToSend, #1

        nop                                      ' SD seems faster than MMC.
        nop                                      ' MMCs need lots of nops sprinkled about.
        nop                                      ' MMCs need at least 3 nops here.
        nop                                      ' (so lets use 5)  
        nop         

        or        outa, clkPinMask               ' clock card hi

        nop
        nop                                      ' MMCs need at least 5 nops here.
        nop                                      ' (so lets use 7)
        nop
        nop
        nop                                      
        nop                                      

        and       outa, !clkPinMask              ' clock card lo 
       
        nop
        nop                                      ' MMCs need at least 4 nops here.
        nop                                      ' (so lets use 6)
        nop
        nop
        nop
         
        djnz      sendIndex, #:loop
        
        or        gotByte,   doneMask            ' set "done" flag
        wrlong    gotByte,   argAddress          ' and return result
    
        jmp       #toploop


argAddress    long $0         
bytToSend     long $0
gotByte       long $0
argData       long $0
temp          long $0   
sendIndex     long $0
       
dataReadyFlag long $0

inPinMask     long $0
outPinMask    long $0
clkPinMask    long $0
readyMask     long $0
doneMask      long $0 

L_MMCSD_CARDIN  long MMCSD_CARDIN  
L_MMCSD_CLK     long MMCSD_CLK
L_MMCSD_CARDOUT long MMCSD_CARDOUT 
          